# Copyright 2022 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import os
from typing import Optional
import pathlib

if False:
    # Used only for typing.
    import stable_baselines3


fruit_slots_home_path = pathlib.Path(os.environ.get('FRUIT_SLOTS_HOME_PATH',
                                                    pathlib.Path.home() / '.fruit_slots'))
log_path: pathlib.Path = fruit_slots_home_path / 'logs'
model_path: pathlib.Path = fruit_slots_home_path / 'models'


def prevent_tensorflow_spam() -> None:
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
    os.environ['CUDA_VISIBLE_DEVICES'] = '0'


def save_model(model: stable_baselines3.PPO, *,
               i_agent: Optional[int] = None, produce_bananas: bool = True,
               produce_lemons: bool = True) -> None:
    agent_path = make_agent_path(i_agent=i_agent,
                                 produce_bananas=produce_bananas, produce_lemons=produce_lemons)
    print(f'Writing model to {agent_path}')
    model.save(agent_path)


def make_agent_path(*, i_agent: Optional[int] = None, produce_bananas: bool = True,
                    produce_lemons: bool = True) -> pathlib.Path:
    result = 'agent-'
    if i_agent is None:
        result += 'single-'
    else:
        result += f'{i_agent}-'

    result += 'a'
    if produce_bananas: result += 'b'
    if produce_lemons: result += 'l'

    result += '.zip'

    return model_path / result


def load_model(*, i_agent: Optional[int] = None, produce_bananas: bool = True,
               produce_lemons: bool = True) -> stable_baselines3.PPO:
    from .fruit_slots_env import FruitSlotsEnv
    import stable_baselines3
    agent_path = make_agent_path(i_agent=i_agent,
                                 produce_bananas=produce_bananas, produce_lemons=produce_lemons)
    if not agent_path.exists():
        raise Exception('You should train before you can use the agents.') from \
                                                                       FileNotFoundError(agent_path)
    print(f'Reading model from {agent_path}')
    return stable_baselines3.PPO.load(
        agent_path,
        env=FruitSlotsEnv.make_and_wrap(is_parallel=False,
                                        produce_bananas=produce_bananas,
                                        produce_lemons=produce_lemons)
    )






# Copyright 2022 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import click


@click.group()
def cli():
    pass

from . import training
from . import playing
from . import plotting


# Copyright 2022 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import random

import numpy as np
import gym.vector.utils
import pettingzoo


N_SLOTS = 10

REWARD_NOTHING = -0.1
REWARD_APPLE = 1
REWARD_BANANA = 5
REWARD_LEMON = -20

EPISODE_LENGTH = 500


class FruitSlotsEnv(pettingzoo.ParallelEnv):

    metadata = {
        "render_modes": ["human"],
        "name": "fruit_slots",
        "is_parallelizable": False,
        "render_fps": 2,
    }

    @staticmethod
    def make_and_wrap(*, produce_bananas=True, produce_lemons=True, is_parallel=False):
        import supersuit as ss
        import stable_baselines3

        env = original_env = FruitSlotsEnv(
            produce_bananas=produce_bananas, produce_lemons=produce_lemons
        )
        env = ss.pettingzoo_env_to_vec_env_v1(env)
        if is_parallel:
            env = ss.concat_vec_envs_v1(env, 8, num_cpus=4, base_class='stable_baselines3')
        else:
            env = ss.concat_vec_envs_v1(env, 1, num_cpus=1, base_class='stable_baselines3')
        env = stable_baselines3.common.vec_env.VecMonitor(
            env,
            info_keywords=(original_env.custom_metrics + ('loggable_metrics',))
        )
        env.original_env = env
        return env


    def __init__(self, *, produce_bananas=True, produce_lemons=True):
        if produce_lemons:
            assert produce_bananas
        self.produce_bananas = produce_bananas
        self.produce_lemons = produce_lemons
        self.possible_agents = ['player_1', 'player_2']
        self._action_space = gym.spaces.Discrete(N_SLOTS)
        self._observation_space = gym.spaces.Box(low=0, high=1, shape=(2, N_SLOTS, 6),
                                                 dtype=bool)
        self.action_spaces = {name: self._action_space for name in self.possible_agents}
        self.observation_spaces = {name: self._observation_space for name in self.possible_agents}
        self.custom_metrics = (
            'cumulative_reward',
            'cumulative_visible_apple_reward',
            'cumulative_invisible_apple_reward',
            *(('cumulative_banana_reward',) if produce_bananas else ()),
            *(('cumulative_lemon_reward',) if produce_lemons else ()),
        )

        self.reset()


    def observation_space(self, agent):
        return self._observation_space

    def action_space(self, agent):
        return self._action_space

    def reset(self, seed=None):
        self.agents = self.possible_agents.copy()

        self._cumulative_rewards = {name: 0 for name in self.agents}
        self._cumulative_visible_apple_rewards = {name: 0 for name in self.agents}
        self._cumulative_invisible_apple_rewards = {name: 0 for name in self.agents}
        self._cumulative_banana_rewards = {name: 0 for name in self.agents}
        self._cumulative_lemon_rewards = {name: 0 for name in self.agents}

        self.agent_locations = {agent: random.randint(0, N_SLOTS - 1) for agent in self.agents}
        self.i_step = 0
        self._remove_all_fruits()
        return self.get_observations()

    def _remove_all_fruits(self):
        self.apple_locations = (set(), set())
        self.visible_apple_locations = (
            (set(), set()),
            (set(), set()),
        )
        self.banana_locations = (set(), set())
        self.lemon_locations = (set(), set())


    def observe(self, agent):
        observation = self._observation_space.low.copy()
        current_agent = agent
        i_current_agent = self.agents.index(current_agent)
        i_other_agent = 1 - i_current_agent

        CHANNEL_STATIC_FALSE = 0
        CHANNEL_STATIC_TRUE = 1
        CHANNEL_AGENT_LOCATIONS = 2
        CHANNEL_APPLE_LOCATIONS = 3
        CHANNEL_BANANA_LOCATIONS = 4
        CHANNEL_LEMON_LOCATIONS = 5

        # Static channels for voodoo reasons:
        observation[:, :, CHANNEL_STATIC_FALSE] = False
        observation[:, :, CHANNEL_STATIC_TRUE] = True

        # Locations of both agents:
        for agent, location in self.agent_locations.items():
            i_agent = self.agents.index(agent)
            observation[i_agent, location, CHANNEL_AGENT_LOCATIONS] = True

        if i_current_agent == 1:
            # Mirror image of agents' locations:
            observation[:, :, CHANNEL_AGENT_LOCATIONS] = \
                                                       observation[::-1, :, CHANNEL_AGENT_LOCATIONS]

        visible_apple_locations_for_current_agent = self.visible_apple_locations[i_current_agent]
        ordered_visible_apple_locations = (visible_apple_locations_for_current_agent if i_agent == 0
                                           else visible_apple_locations_for_current_agent[::-1])
        for i, visible_apple_locations in enumerate(ordered_visible_apple_locations):
            for visible_apple_location in visible_apple_locations:
                observation[i, visible_apple_location, CHANNEL_APPLE_LOCATIONS] = True

        if self.banana_locations[i_other_agent]:
            (banana_location,) = self.banana_locations[i_other_agent]
            observation[1, banana_location, CHANNEL_BANANA_LOCATIONS] = True

        for lemon_location in self.lemon_locations[i_other_agent]:
            observation[1, lemon_location, CHANNEL_LEMON_LOCATIONS] = True

        assert observation.shape == self._observation_space.shape
        return observation


    def get_observations(self):
        return {agent: self.observe(agent) for agent in self.agents}


    def step(self, actions):
        rewards = {agent: 0 for agent in self.agents}
        infos = {agent: {} for agent in self.agents}
        dones = {agent: self.i_step >= 500 for agent in self.agents}
        possible_new_fruit_locations = tuple(set(range(N_SLOTS)) - set(self.agent_locations)
                                             - set(actions.values()))
        self.agent_locations.update(actions)

        for i_current_agent, (current_agent, action) in enumerate(actions.items()):
            i_other_agent = 1 - i_current_agent
            other_agent = self.agents[i_other_agent]

            ### Dealing with agent eating apple: ###################################################
            #                                                                                      #
            apple_locations_for_current_agent = self.apple_locations[i_current_agent]
            try:
                apple_locations_for_current_agent.remove(action)
            except KeyError:
                pass
            else:
                # Agent ate an apple.
                rewards[current_agent] += REWARD_APPLE

                try:
                    self.visible_apple_locations[i_current_agent][i_current_agent].remove(action)
                except KeyError:
                    # Agent ate an invisible apple.
                    self._cumulative_invisible_apple_rewards[current_agent] += REWARD_APPLE
                    self.visible_apple_locations[i_other_agent][i_current_agent].remove(action)
                else:
                    # Agent ate a visible apple.
                    assert (action not in
                            self.visible_apple_locations[i_other_agent][i_current_agent])
                    self._cumulative_visible_apple_rewards[current_agent] += REWARD_APPLE
            #                                                                                      #
            ### Finished dealing with agent eating apple. ##########################################

            ### Dealing with agent eating banana: ##################################################
            #                                                                                      #
            try:
                self.banana_locations[i_current_agent].remove(action)
            except KeyError:
                pass
            else:
                # Agent ate a banana.
                rewards[current_agent] += REWARD_BANANA
                rewards[other_agent] += REWARD_BANANA
                self._cumulative_banana_rewards[current_agent] += REWARD_BANANA
                self._cumulative_banana_rewards[other_agent] += REWARD_BANANA
            #                                                                                      #
            ### Finished dealing with agent eating banana. #########################################


            ### Dealing with agent eating lemon: ###################################################
            #                                                                                      #

            try:
                self.lemon_locations[i_current_agent].remove(action)
            except KeyError:
                pass
            else:
                # Agent ate a lemon.
                rewards[current_agent] = REWARD_NOTHING
                rewards[other_agent] = REWARD_LEMON
                self._cumulative_lemon_rewards[other_agent] += REWARD_LEMON

            #                                                                                      #
            ### Finished dealing with agent eating lemon. ##########################################

            if rewards[current_agent] == 0:
                rewards[current_agent] = REWARD_NOTHING

            self._cumulative_rewards[current_agent] += rewards[current_agent]

            infos[current_agent] = {
                metric: getattr(self, f'_{metric}s')[current_agent]
                for metric in self.custom_metrics
            }
            infos[current_agent]['loggable_metrics'] = self.custom_metrics

            self.agent_locations[current_agent] = action

        ### Advancing turn and dealing with scheduled events: ######################################
        #                                                                                          #
        self.i_step += 1

        if self.i_step % 5 == 0:
            self._remove_all_fruits()

            if self.produce_lemons and self.i_step % 100 == 0:
                i_agent_on_banana_side, i_agent_on_lemon_side = random.sample(range(2), 2)
                new_banana_location, *new_lemon_locations = \
                                                      random.sample(possible_new_fruit_locations, 6)
                self.banana_locations[i_agent_on_banana_side].add(new_banana_location)
                self.lemon_locations[i_agent_on_lemon_side].update(new_lemon_locations)

            elif self.produce_bananas and self.i_step % 25 == 0:
                new_banana_location = random.choice(possible_new_fruit_locations)
                random.choice(self.banana_locations).add(new_banana_location)

            elif self.i_step % 5 == 0:
                new_apple_location = random.choice(possible_new_fruit_locations)
                for apple_locations in self.apple_locations:
                    apple_locations.add(new_apple_location)
                i_agent_that_can_see_new_apple_pair = random.randint(0, 1)
                for i_agent in range(2):
                    self.visible_apple_locations[i_agent_that_can_see_new_apple_pair][i_agent]. \
                                                                             add(new_apple_location)
        #                                                                                          #
        ### Finished advancing turn and dealing with scheduled events. #############################

        return self.get_observations(), rewards, dones, infos


    def render(self, mode='human'):
        result = [([' '] * N_SLOTS) for i_agent in range(2)]
        for agent, agent_location in self.agent_locations.items():
            i_agent = self.agents.index(agent)
            result[i_agent][agent_location] = str(i_agent + 1)
        for i_agent, apple_locations in enumerate(self.apple_locations):
            for apple_location in apple_locations:
                apple_pair_is_visible = (apple_location in
                                         self.visible_apple_locations[i_agent][i_agent])
                result[i_agent][apple_location] = ('A' if apple_pair_is_visible else 'a')
        for i_agent, banana_locations in enumerate(self.banana_locations):
            if banana_locations:
                (banana_location,) = banana_locations
                result[i_agent][banana_location] = 'B'

        for i_agent, lemon_locations in enumerate(self.lemon_locations):
            for lemon_location in lemon_locations:
                result[i_agent][lemon_location] = 'L'

        top_horizontal_line = '/' + '-' * N_SLOTS + '\\'
        bottom_horizontal_line = '\\' + '-' * N_SLOTS + '/'
        return (top_horizontal_line + '\n|' + ''.join(result[0]) + '|\n|' + ''.join(result[1]) +
                '|\n' + bottom_horizontal_line)


    def close(self):
        pass



import random
import os
import logging
import sys
import itertools
import warnings
import functools

import pathlib

import click
import pandas as pd

from fruit_slots import utils
from . import cli



def get_dataframe(tensorboard_log_path):
    utils.prevent_tensorflow_spam()
    import tensorflow.compat.v1.logging
    tensorflow.compat.v1.logging.set_verbosity(tensorflow.compat.v1.logging.ERROR)
    from tensorflow.python.summary.summary_iterator import summary_iterator

    events = tuple(summary_iterator(str(tensorboard_log_path)))
    d = {}
    columns = {'step'}
    for event in events:
        try:
            step_d = d[event.step]
        except KeyError:
            step_d = d[event.step] = {'step': event.step}
        for value in event.summary.value:
            columns.add(value.tag)
            step_d[value.tag] = value.simple_value
    records = sorted(d.values(), key=lambda step_d: step_d['step'])
    df = pd.DataFrame.from_records(records, columns=columns, index='step')
    return df


all_columns = (
    'rollout/mean_cumulative_visible_apple_reward',
    'rollout/mean_cumulative_invisible_apple_reward',
    'rollout/mean_cumulative_banana_reward',
    'rollout/mean_cumulative_lemon_reward',
)


@cli.command()
@click.argument('tensorboard_log_path', type=str, default='')
def make_chart(tensorboard_log_path):
    import plotly.graph_objects as go
    if tensorboard_log_path == '':
        recent_tensorboard_log_folder = max(utils.log_path.iterdir(),
                                            key=lambda path: path.stat().st_mtime)
        (tensorboard_log_path,) = recent_tensorboard_log_folder.iterdir()
    else:
        tensorboard_log_path = pathlib.Path(tensorboard_log_path)
    print(f'Making a chart for {tensorboard_log_path}')
    df = get_dataframe(tensorboard_log_path)
    columns = tuple(column for column in all_columns if column in df.columns)
    process_column_name = lambda name: (
        name.replace('rollout/mean_cumulative_', '').replace('_', ' ').capitalize()
    )
    axis_template = {'title_font': {'size': 23,}, 'tickfont': {'size': 20,}}
    figure = go.Figure(
        data=[
            go.Scatter(x=df.index, y=df[column], name=process_column_name(column))
            for column in columns
        ],
        layout=go.Layout(
            title=('Points gained over generations of training PPO on Fruit Slots, single '
                   'neural network'),
            title_font_size=28,
            xaxis={**axis_template, 'title': 'Generation'},
            yaxis={**axis_template, 'title': 'Points'},
            legend={'xanchor': 'right', 'x': 1, 'yanchor': 'bottom',
                    'y': 0, 'font': {'size': 30}}
        )
    )
    figure.show()





# Copyright 2022 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import random
import sys
import itertools
import warnings
import functools

import unittest.mock
import pathlib

import click
import numpy as np

from fruit_slots import utils
from . import cli

if False:
    # Used only for typing.
    import stable_baselines3


@cli.command()
@click.option('-p', '--parallel', 'is_parallel', default=False)
@click.option('--bananas/--no-bananas', 'produce_bananas', default=True)
@click.option('--lemons/--no-lemons', 'produce_lemons', default=True)
@click.option('-t', '--total-timesteps', default=1_000_000)
@click.option('-v', '--verbose', default=False, is_flag=True)
def train_single(*, is_parallel, produce_bananas, produce_lemons, total_timesteps, verbose):
    import stable_baselines3
    from fruit_slots.fruit_slots_env import FruitSlotsEnv
    utils.prevent_tensorflow_spam()

    class TensorboardCallback(stable_baselines3.common.callbacks.BaseCallback):
        def __init__(self, verbose=0):
            super().__init__(verbose=verbose)

        def _on_step(self):
            from stable_baselines3.common.utils import safe_mean

            log_interval = self.locals['log_interval']
            iteration = self.locals['iteration']
            model = self.locals['self']
            ep_info_buffer = model.ep_info_buffer
            if (log_interval is not None and iteration % log_interval == 0 and
                len(ep_info_buffer) > 0 and len(ep_info_buffer[0]) > 0):

                for metric_name in ep_info_buffer[0].get('loggable_metrics', ()):
                    self.logger.record(
                        f'rollout/mean_{metric_name}',
                        safe_mean([ep_info[metric_name] for ep_info in ep_info_buffer])
                    )
            return True

    env = FruitSlotsEnv.make_and_wrap(is_parallel=is_parallel,
                                      produce_bananas=produce_bananas,
                                      produce_lemons=produce_lemons)
    model = stable_baselines3.PPO(stable_baselines3.ppo.MlpPolicy, env, n_steps=32,
                                  tensorboard_log=utils.log_path, verbose=verbose)

    print('Starting learning... ')
    model.learn(total_timesteps=total_timesteps, n_eval_episodes=30, eval_freq=1,
                callback=TensorboardCallback())
    print('Done learning.')
    utils.save_model(model, produce_bananas=produce_bananas, produce_lemons=produce_lemons)



# Copyright 2022 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import random
import sys
import itertools
import warnings
import functools

import unittest.mock
import pathlib

import click
import numpy as np

from fruit_slots import utils
from . import cli


@cli.command()
@click.option('--bananas/--no-bananas', 'produce_bananas', default=True)
@click.option('--lemons/--no-lemons', 'produce_lemons', default=True)
def play_single(*, produce_bananas, produce_lemons):
    from fruit_slots.fruit_slots_env import FruitSlotsEnv
    env = FruitSlotsEnv(produce_bananas=produce_bananas, produce_lemons=produce_lemons)
    model = utils.load_model(produce_bananas=produce_bananas, produce_lemons=produce_lemons)
    print('Starting playing... ')
    observations = env.reset()
    print(env.render())
    print()
    dones = {agent: False for agent in env.agents}
    while False in dones.values():
        actions = {agent: model.predict(observation, deterministic=True)[0]
                   for agent, observation in observations.items()}
        observations, rewards, dones, infos = env.step(actions)
        print(f'Actions: {actions["player_1"]} / {actions["player_2"]}')
        print(env.render())
        print(f'Rewards: {rewards["player_1"]} / {rewards["player_2"]}')
        print()
    print('Done playing.')





# Copyright 2022 Ram Rachum and collaborators.
# This program is distributed under the MIT license.
import setuptools
import re


def read_file(filename):
    with open(filename) as file:
        return file.read()

version = re.search("__version__ = '([0-9.]*)'",
                    read_file('fruit_slots/__init__.py')).group(1)

setuptools.setup(
    name='fruit_slots',
    version=version,
    author='Ram Rachum',
    author_email='ram@rachum.com',
    description='An experiment in implicit communication between learning agents',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/cool-RR/fruit_slots',
    packages=setuptools.find_packages(exclude=['tests*']),
    install_requires=read_file('requirements.txt'),
    extras_require={
        'tests': {
            'pytest',
        },
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)
